#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2016 Century Systems Co.,Ltd.
# Last Modified: 2016/04/25 14:20:28 kikuchi

import struct
import collections

from lib import protocol
from lib import hdlc


#=====================================================================
#
#
#
#=====================================================================
class OKI_API(object):
    #-------------------------------------------------------
    #
    #-------------------------------------------------------
    def __init__(self, port):
        self._oki = hdlc.OKI_Serial(port)

        self._s_fb_header = struct.Struct(">BH")
        self._s_api_header = struct.Struct(">BHBBHH")

        self._seq = 0
        self._callback = None

        self._data_queue = []

    #-------------------------------------------------------
    #
    #-------------------------------------------------------
    def fileno(self):
        return self._oki.fileno()

    #-------------------------------------------------------
    #
    #-------------------------------------------------------
    def _inc_seq(self):
        self._seq = (self._seq + 1) & 0xffff

    #-------------------------------------------------------
    #
    #-------------------------------------------------------
    def _make_API_frame(self, msgid, data):
        frame = self._s_api_header.pack(0x10, 0x8f8f, 0x01, msgid,
                                        self._seq, len(data)) + data
        self._inc_seq()

        return frame

    #-------------------------------------------------------
    #
    #-------------------------------------------------------
    def _send_receive_API(self, buf, timeout=None):
        written = self._oki.send_API(buf)

        while True:
            res = self._oki.receive(timeout)

            if not res:
                break

            if res["ProcessID"] == hdlc.ProcessID.API:
                break
            elif res["ProcessID"] == hdlc.ProcessID.DATA:
                self._data_queue.append(res["payload"]["data"])

        return res

    #-------------------------------------------------------
    #
    #-------------------------------------------------------
    def register_data_callback(self, function, arg):
        self._callback = {"function" : function,
                          "arg"      : arg}

    #-------------------------------------------------------
    # OKI API: [msgid: 0x0C] リセット要求
    #-------------------------------------------------------
    def API_reset(self):
        data = protocol.make_cmd0c_cmd()
        frame = self._make_API_frame(0x0c, data)

        while True:
            res = self._send_receive_API(frame)

            if not res:
                break

            if res["payload"]["cmdid"] == 2:
                break

        return self.API_wait_boot(5)

    #-------------------------------------------------------
    # OKI API: 起動完了通知待ち
    #-------------------------------------------------------
    def API_wait_boot(self, timeout=None):
        last_data = None

        while True:
            data = self._oki.receive(timeout)

            if not data:
                break

            last_data = data
            timeout = 1

        res = False
        if last_data:
            payload = last_data["payload"]

            if payload["cmdid"] == 3 and payload["msgid"] == 0xd2:
                res = True

        return res

    #-------------------------------------------------------
    # OKI API: [msgid: 0x01] 基本設定要求
    #-------------------------------------------------------
    def API_set_basic_setting(self, pan_id, nw_id, ch):
        data = protocol.make_cmd01_set_cmd(pan_id, nw_id, ch)
        frame = self._make_API_frame(0x01, data)

        while True:
            res = self._send_receive_API(frame)

            if not res:
                break

            if res["payload"]["cmdid"] == 2:
                break

        decoded = protocol.decode_cmd01_set_res(res["payload"]["data"])
        res["response"] = decoded

        if decoded.req == 0x00:
            ret = True
        else:
            ret = False

        return ret, res

    #-------------------------------------------------------
    # OKI API: [msgid: 0x01] 基本設定取得要求
    #-------------------------------------------------------
    def API_get_basic_setting(self):
        data = protocol.make_cmd01_get_cmd()
        frame = self._make_API_frame(0x01, data)

        while True:
            res = self._send_receive_API(frame)

            if not res:
                break

            if res["payload"]["cmdid"] == 2:
                break

        decoded = protocol.decode_cmd01_get_res(res["payload"]["data"])
        res["response"] = decoded

        if decoded.req == 0x00:
            ret = True
        else:
            ret = False

        return ret, res

    #-------------------------------------------------------
    # OKI API: [msgid: 0x06] 局番号解析用データ設定
    #-------------------------------------------------------
    def API_set_protocol_setting(self):
        data = protocol.make_cmd06_set_cmd()
        frame = self._make_API_frame(0x06, data)

        while True:
            res = self._send_receive_API(frame)

            if not res:
                break

            if res["payload"]["cmdid"] == 2:
                break

        decoded = protocol.decode_cmd06_set_res(res["payload"]["data"])
        res["response"] = decoded

        if decoded.req == 0x00:
            ret = True
        else:
            ret = False

        return ret, res

    #-------------------------------------------------------
    # OKI API: [msgid: 0x08] 経路情報取得要求
    #-------------------------------------------------------
    def API_get_route_info(self, mac=0):
        data = protocol.make_cmd08_get_cmd(2, mac)
        frame = self._make_API_frame(0x08, data)

        while True:
            res = self._send_receive_API(frame)

            if not res:
                break

            if res["payload"]["cmdid"] == 2:
                break

        decoded = protocol.decode_cmd08_get_res(res["payload"]["data"])
        res["response"] = decoded
        res["req"] = decoded.req

        if decoded.req == 0x00:
            res["end"] = decoded.end
            res["num"] = decoded.num
            res["total"] = decoded.total
            res["dodag"] = decoded.dodag
            if hasattr(decoded, "global_"):
                ret = True
                res["global"] = decoded.global_
                res["parent"] = decoded.parent
                res["mac"] = decoded.mac
                res["parentmac"] = decoded.parentmac
            else:
                ret = False
        else:
            ret = False

        return ret, res

    #-------------------------------------------------------
    # API: [msgid: 0x10] 暗号化鍵登録要求
    #-------------------------------------------------------
    def API_set_nwkey_setting(self, nwkey=None):
        data = protocol.make_cmd10_set_cmd(nwkey)
        frame = self._make_API_frame(0x10, data)

        while True:
            res = self._send_receive_API(frame)

            if not res:
                break

            if res["payload"]["cmdid"] == 2:
                break

        decoded = protocol.decode_cmd10_set_res(res["payload"]["data"])
        res["response"] = decoded

        if decoded.req == 0x00:
            ret = True
        else:
            ret = False

        return ret, res

    #-------------------------------------------------------
    # API: [msgid: 0x1A] モジュール状態情報取得要求
    #-------------------------------------------------------
    def API_get_module_status(self, typ):
        data = protocol.make_cmd1A_set_cmd(typ)
        frame = self._make_API_frame(0x1A, data)

        while True:
            res = self._send_receive_API(frame)

            if not res:
                break

            if res["payload"]["cmdid"] == 2:
                break

        decoded = protocol.decode_cmd1A_set_res(res["payload"]["data"])
        if decoded:
            res["response"] = decoded

            if decoded.req == 0x00:
                ret = True
            else:
                ret = False
        else:
            ret = False

        return ret, res

    #-------------------------------------------------------
    # API: [msgid: 0x21] ノード生存確認要求
    #-------------------------------------------------------
    def API_send_ping(self, ipaddress, timeout=None):
        data = protocol.make_cmd21_send_cmd(ipaddress)
        frame = self._make_API_frame(0x21, data)

        while True:
            res = self._send_receive_API(frame, timeout)

            if not res:
                break

            if res["payload"]["cmdid"] == 2:
                break

        if not res:
            return False, res

        decoded = protocol.decode_cmd21_send_res(res["payload"]["data"])
        res["response"] = decoded

        if decoded.req == 0x00:
            ret = True
        else:
            ret = False

        return ret, res

    #-------------------------------------------------------
    # API: [msgid: 0x23] パケットフィルタリング機能設定
    #-------------------------------------------------------
    def API_set_packet_filtering(self, Pfilter, Timeout=10, Resptime=0):
        data = protocol.make_cmd23_set_cmd(Pfilter, Timeout, Resptime)
        frame = self._make_API_frame(0x23, data)

        while True:
            res = self._send_receive_API(frame)

            if not res:
                break

            if res["payload"]["cmdid"] == 2:
                break

        decoded = protocol.decode_cmd23_set_res(res["payload"]["data"])
        res["response"] = decoded

        if decoded.req == 0x00:
            ret = True
        else:
            ret = False

        return ret, res

    #-------------------------------------------------------
    # [msgid: 0x88] 局番号リスト設定 (lowlevel)
    #-------------------------------------------------------
    def _set_addrlist_setting(self, page, addr_list):
        data = protocol.make_cmd88_cmd(page, addr_list)
        frame = self._make_API_frame(0x88, data)

        while True:
            res = self._send_receive_API(frame)

            if not res:
                break

            if res["payload"]["cmdid"] == 2:
                break

        decoded = protocol.decode_cmd88_set_res(res["payload"]["data"])
        res["response"] = decoded

        if decoded.req == 0x00:
            ret = True
        else:
            ret = False

        return ret, res

    #-------------------------------------------------------
    # OKI API: [msgid: 0x88] 局番号リスト設定
    #-------------------------------------------------------
    def API_set_addrlist_setting(self, addr_list):
        if len(addr_list) > 300:
            return False

        splitted = [addr_list[:150], addr_list[150:]]

        responses = []

        result = True

        for page in range(2):
            ret, res = self._set_addrlist_setting(page+1, splitted[page])

            responses.append(res)

            if ret == False:
                result = False

        return result, responses

    #-------------------------------------------------------
    # OKI API: [msgid: 0xD3] 運用開始要求
    #-------------------------------------------------------
    def API_start_operation(self):
        frame = self._make_API_frame(0xD3, b"")

        while True:
            res = self._send_receive_API(frame)

            if not res:
                break

            if res["payload"]["cmdid"] == 2:
                break

        decoded = protocol.decode_cmdD3_set_res(res["payload"]["data"])
        res["response"] = decoded

        if decoded.req == 0x00:
            ret = True
        else:
            ret = False

        return ret, res

    #-------------------------------------------------------
    # OKI API: 透過データ write
    #-------------------------------------------------------
    def write_transparent(self, data):
        ret = self._oki.send_transparent(data)

        return ret

    #-------------------------------------------------------
    # OKI API: 透過データ read
    #-------------------------------------------------------
    def read_transparent(self, timeout=None):
        if len(self._data_queue) > 0:
            data = self._data_queue.pop(0)
            return data

        while True:
            res = self._oki.receive(timeout)

            if not res:
                return None
            elif res["ProcessID"] == hdlc.ProcessID.DATA:
                return res["payload"]["data"]
