#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2016 Century Systems Co.,Ltd.
# Last Modified: 2016/04/20 11:24:36 kikuchi

import enum
import serial
import select
import struct

ESC = b'\x7d'
SYN = b'\x7e'
XOR = b'\x20'


class ProcessID(enum.Enum):
    API     = 0x10    # ホストCPUとモジュール間のAPI
    DATA    = 0x11    # ホストCPUとモジュール間の透過データ
    MEASURE = 0x12    # ホストCPU間の測定モード制御用API


class cmdid(enum.Enum):
    Request  = 0x01
    Response = 0x02
    Notice   = 0x03


#=====================================================================
#
#
#
#=====================================================================
class OKI_Serial(object):
    #-------------------------------------------------------
    #
    #-------------------------------------------------------
    def __init__(self, port):
        self._s = serial.Serial(port, baudrate=115200, timeout=0.1)

        # struct to parse response
        self._s_fb_header = struct.Struct(">BH")
        self._s_api_header = struct.Struct(">BHBBHH")

    #-------------------------------------------------------
    #
    #-------------------------------------------------------
    def fileno(self):
        return self._s.fileno()

    #-------------------------------------------------------
    #
    #-------------------------------------------------------
    def _read_frame(self, initial_timeout=None):
        rfds, _, _ = select.select([self._s.fileno()], [], [], initial_timeout)

        if not rfds:
            return None

        line = bytearray()
        delimiter = SYN
        in_frame = False

        while True:
            c = self._s.read(1)

            if c:
                if in_frame == True:
                    if c != delimiter:
                        line += c
                    elif len(line) == 0:
                        # sync error -> recover
                        in_frame = False
                    else:
                        break
                elif c == delimiter:
                    in_frame = True
            else:
                # framing error
                return None

        return bytes(line)

    #-------------------------------------------------------
    #
    #-------------------------------------------------------
    def _escape_frame(self, frame, with_SYN=False):
        if with_SYN:
            l = [SYN]
        else:
            l = []

        for c in frame:
            if c in (ord(ESC), ord(SYN)):
                l.append(ESC)
                c = c ^ ord(XOR)

            l.append(chr(c).encode('latin1'))

        if with_SYN:
            l.append(SYN)

        buf = b''.join(l)
        return buf

    #-------------------------------------------------------
    #
    #-------------------------------------------------------
    def _unescape_frame(self, frame):
        l = []

        unescape_next = False

        for c in frame:
            if c == ord(ESC):
                unescape_next = True
            else:
                if unescape_next == True:
                    c = c ^ ord(XOR)
                    unescape_next = False

                l.append(chr(c).encode('latin1'))

        buf = b''.join(l)
        return buf

    #-------------------------------------------------------
    #
    #-------------------------------------------------------
    def _make_frame(self, process_id, data):
        frame = struct.pack(">BH", process_id, len(data)) + data
        buf = self._escape_frame(frame, with_SYN=True)

        return buf

    #-------------------------------------------------------
    #
    #-------------------------------------------------------
    def _send_data(self, process_id, data):
        buf = self._make_frame(process_id, data)

        return self._s.write(buf)

    #-------------------------------------------------------
    # API: APIデータ送信
    #-------------------------------------------------------
    def send_API(self, data):
        return self._send_data(0x10, data)

    #-------------------------------------------------------
    # API: 透過データ送信
    #-------------------------------------------------------
    def send_transparent(self, data):
        return self._send_data(0x11, data)
 
    #-------------------------------------------------------
    # API: 受信
    #-------------------------------------------------------
    def receive(self, timeout=None):
        buf = self._read_frame(initial_timeout=timeout)

        if not buf:
            return None

        buf = self._unescape_frame(buf)

        fb_header_len = self._s_fb_header.size
        fb_header = self._s_fb_header.unpack(buf[:fb_header_len])
        # -> (ProcessID, Length)

        payload_len = fb_header[1]

        if fb_header[0] == ProcessID.API.value:
            api_header_len = self._s_api_header.size
            api_header = self._s_api_header.unpack(
                buf[fb_header_len:fb_header_len+api_header_len])
            # -> (id, auth, cmdid, msgid, seq, len)

            data = buf[fb_header_len+api_header_len:]

            res = {"ProcessID" : ProcessID.API,
                   "Length"    : fb_header[1],
                   "payload" : {
                       "id"    : api_header[0],
                       "auth"  : api_header[1],
                       "cmdid" : api_header[2],
                       "msgid" : api_header[3],
                       "seq"   : api_header[4],
                       "data"  : data
                    }
            }
        elif fb_header[0] == ProcessID.DATA.value:
            data = buf[fb_header_len:]

            res = {"ProcessID" : ProcessID.DATA,
                   "Length"    : fb_header[1],
                   "payload" : {
                       "data"  : data
                    }
            }
        else:
            res = None

        return res
