#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2016 Century Systems Co.,Ltd.
# Last Modified: 2016/04/25 16:48:01 kikuchi

# import json
import configparser
import select
import os
import stat
import pty
import re
import tty
import shutil
import signal
import sys
import syslog
from datetime import datetime
import time

from lib import api


#=====================================================================
#
#
#
#=====================================================================
class MH920Bridge(object):
    #-------------------------------------------------------
    #
    #-------------------------------------------------------
    def __init__(self, conf_file_name):
        conf_file = configparser.ConfigParser()
        with open(conf_file_name, 'r') as f:
            conf_file.read_file(f)

        conf = {}

        # require section (if not exist, raise Exception)
        general = conf_file['General']

        # require
        conf['pan_id'] = general.getint('pan_id')
        # TODO: range

        conf['nw_id'] = general.get('nw_id')
        # TODO: format check

        conf['channel'] = general.getint('channel')
        # TODO: range

        # option
        conf['device'] = general.get('device', '/dev/ttyS0')
        # TODO: format check

        conf['virtdev_name'] = general.get('virtdev', '/dev/mh920')
        # TODO: format check

        nwkey = general.get('nwkey')
        if nwkey:
            conf['nwkey'] = bytes.fromhex(nwkey)
        else:
            conf['nwkey'] = False
        # TODO: range

        conf['debug'] = general.getboolean('debug', False)

        conf['init_reset'] = general.getboolean('init_reset', True)

        pfilter = {}
        if 'PFilter' in conf_file:
            pfilter_file = conf_file['PFilter']

            pfilter['pfilter'] = pfilter_file.getint('pfilter', 1)
            # TODO: range

            timeout = pfilter_file.getfloat('timeout', 4)
            pfilter['timeout'] = int(timeout * 10)
            # TODO: range
        else:
            pfilter['pfilter'] = 1
            pfilter['timeout'] = 4 * 10

        conf['pfilter'] = pfilter

        ping_mode_file = conf_file['PingMode']

        conf['ping_mode'] = ping_mode_file.getboolean('enable', False)
        conf['ping_interval'] = ping_mode_file.getfloat('interval', 10)
        conf['ping_count'] = ping_mode_file.getint('count', 10)

        ping_mac_list = []
        for name, value in ping_mode_file.items():
            if name.startswith('mac_'):
                ping_mac_list.append(value)
        conf['ping_mac_list'] = ping_mac_list

        addr_list_file = conf_file['ShortAddress']
        addr_list = []
        for name, value in addr_list_file.items():
            name_int = int(name, 0)
            value_int = int(value, 0)
            t = name_int, value_int
            addr_list.append(t)
        conf['addr_list'] = addr_list

        self._conf = conf

        self._initialized = False

    #-------------------------------------------------------
    #
    #-------------------------------------------------------
    def _sig_handler(self, signum, frame):
        msg = "Signal No.{} received.".format(signum)
        syslog.syslog(msg)
        raise Exception(msg)

    #-------------------------------------------------------
    #
    #-------------------------------------------------------
    def _print(self, *args, **kwargs):
        if self.getconf('debug') == True:
            print(*args, **kwargs)

    #-------------------------------------------------------
    #
    #-------------------------------------------------------
    def getconfdic(self):
        return self._conf

    def getconf(self, key):
        return self._conf[key]

    def dumpconf(self):
        conf = self.getconfdic()
        print(str(conf))

    def checkconf(self):
        '''内部用チェックルーチン。_confに必要な設定がされているか調べる'''
        conf_keys=['pan_id', 'nw_id', 'channel', 'device', 'virtdev_name',
            'nwkey', 'debug', 'pfilter', 'ping_mode', 'ping_interval',
            'ping_mac_list', 'addr_list']

        for key in conf_keys:
            if key not in self._conf:
                raise Exception("No key " + key + " in config.")

    #-------------------------------------------------------
    #
    #-------------------------------------------------------
    def init(self):
        conf = self.getconfdic()

        # okiをセット
        self._oki = api.OKI_API(conf['device'])

        if conf['init_reset']:
            # リセット要求
            self._print("# API_reset:")
            res = self._oki.API_reset()

            if not res:
                return False

        # 基本設定要求
        self._print("# API_set_basic_setting:")
        res, _ = self._oki.API_set_basic_setting(
            conf['pan_id'], conf['nw_id'], conf['channel']
        )

        if not res:
            return False

        # 局番号解析用データ設定 (currently Modbus-RTU only)
        self._print("# API_set_protocol_setting:")
        res, _ = self._oki.API_set_protocol_setting()

        if not res:
            return False

        # 暗号化鍵登録要求
        if conf['nwkey']:
            self._print("# API_set_nwkey_setting:")
            res, _ = self._oki.API_set_nwkey_setting(conf['nwkey'])

            if not res:
                return False

        # 局番号リスト設定
        self._print("# API_set_addrlist_setting:")
        res, _ = self._oki.API_set_addrlist_setting(conf['addr_list'])

        if not res:
            return False

        # パケットフィルタリング機能設定
        pfilter = conf['pfilter']
        self._print("# API_set_packet_filtering:")
        res, _ = self._oki.API_set_packet_filtering(
            pfilter['pfilter'], pfilter['timeout'], 0
        )

        if not res:
            return False

        # 運用開始要求
        self._print("# API_start_operation:")
        res, _ = self._oki.API_start_operation()

        if not res:
            return False

        self._initialized = True

        return True

    #-------------------------------------------------------
    #
    #-------------------------------------------------------
    def _init_pty(self):
        ptys = pty.openpty()

        tty.setraw(ptys[0])

        devname = self.getconf('virtdev_name')
        ttyname = os.ttyname(ptys[1])

        shutil.chown(ttyname, "root", "dialout")
        os.chmod(ttyname, stat.S_IRUSR|stat.S_IWUSR|stat.S_IRGRP|stat.S_IWGRP)
        os.symlink(ttyname, devname)

        self._pty_fd = ptys[0]
        self._devname = devname

        signal.signal(signal.SIGTERM, self._sig_handler)
        signal.signal(signal.SIGINT, self._sig_handler)

    #-------------------------------------------------------
    #
    #-------------------------------------------------------
    def mainloop(self):
        if not self._initialized:
            return False

        self._init_pty()

        oki_fd = self._oki.fileno()

        try:
            while True:
                rlist = [oki_fd, self._pty_fd]
                rfds, _, _ = select.select(rlist, [], [])

                if oki_fd in rfds:
                    # read from MH-920, write to pty
                    buf = self._oki.read_transparent()
                    if buf:
                        self._print("* read from MH920, {}".format(buf))
                        os.write(self._pty_fd, buf)
                elif self._pty_fd in rfds:
                    # read from pty, write to MH-920
                    buf = os.read(self._pty_fd, 1500)
                    if buf:
                        self._print("# read from pty, {}".format(buf))
                        res = self._oki.write_transparent(buf)
                        self._print(" -> res: {}".format(res))
        except Exception as err:
            self._print("mainloop caught exception, '{}'".format(err))
            os.unlink(self._devname)

    def pingloop(self):
        if not self._initialized:
            return False

        count = self.getconf('ping_count')
        init_count = 5
        result = {}
        try:
            for i in range(init_count):
                ret, res = self._oki.API_get_module_status(1)
                if ret == True:
                    if res['response'].actled == 0x02:
                        print('confirmed joinning network.')
                        break
                    else:
                        print('actled = {}'.format(res['response'].actled))
                print("wait for joinning network.")
                time.sleep(10)
            else:
                print('cannot join network. exit.')
                return False

            # MACアドレスに対応するIPアドレスを取得
            self._print('Getting route info.')
            ipaddrlist = []
            maclist = self.getconf('ping_mac_list')
            for mac in maclist:
                macb = bytes.fromhex(mac)
                for i in range(count):
                    ret, res = self._oki.API_get_route_info(macb)
                    if (ret == True):
                        t = (mac, res['global'])
                        ipaddrlist.append(t)
                        break
                    time.sleep(5)
                else:
                    print('cannot get IP address for ' + mac )

            # IPアドレスに対してノード生存確認要求を送る
            ping_interval = self.getconf('ping_interval')

            for mac, ip in ipaddrlist:
                result[mac] = [0, 0]

            for c in range(count):
                for mac, ip in ipaddrlist:
                    now = datetime.now()
                    start = time.monotonic()
                    ret, res = self._oki.API_send_ping(ip, ping_interval)
                    end = time.monotonic()
                    timestr = now.strftime('[%H:%M:%S] ')
                    if (ret == True):
                        resultstr = 'time={:.3f} ms'.format((end-start)*1000)
                        result[mac][0] += 1
                    else:
                        resultstr = 'No response.'
                        result[mac][1] += 1
                    print(timestr + 'From "' + mac + '" ' + resultstr)
                time.sleep(ping_interval)
        except Exception as err:
            self._print("pingloop caught exception, '{}'".format(err))

        #結果集計
        for mac in result:
            ok = result[mac][0]
            ng = result[mac][1]
            print(mac + ": OK={OK}, NG={NG}".format(OK=ok, NG=ng))

def main():
    if (len(sys.argv) == 1):
        conf_file = 'mh920_bridge.ini'
    else:
        conf_file = sys.argv[1]

    bridge = MH920Bridge(conf_file)

    bridge.checkconf()
    # bridge.dumpconf()

    f = bridge.init()
    if f == False:
        print("init failed.")
        return

    if (bridge.getconf('ping_mode')):
        bridge.pingloop()
    else:
        bridge.mainloop()

if __name__ == "__main__":
    main()
